import numpy as np
import seaborn as sns
from torchvision.transforms.functional import to_pil_image
from . import load, infer, process, measure, draw

def two_points_linear_constant(a, b):
    aX = a[0]
    aY = a[1]
    bX = b[0]
    bY = b[1]
    if (bX-aX) != 0:
        alpha = (bY - aY)/(bX - aX)
        beta = (bX * aY - bY * aX)/(bX - aX)
    else:
        alpha = None
        beta = None
    return alpha, beta


def find_segment_line(segment_area, alpha, beta):
    line_coordinates = []
    for idx, position in enumerate(sorted(segment_area)):
        expectedY = alpha * position[0] + beta
        if position[1] == int(expectedY):
            line_coordinates.append(
                [position[0], position[1]]
            )
    return line_coordinates


def SHA_score(ls, rs, lh, rh, la, ra):
    sL = measure.two_points_distance(ls, rs)
    hL = measure.two_points_distance(lh, rh)
    ms = measure.find_middle_point(ls, rs)
    mh = measure.find_middle_point(lh, rh)
    ma = measure.find_middle_point(la, ra)
    bL = measure.two_points_distance(ms, mh)
    lL = measure.two_points_distance(mh, ma)
    score = int((hL * lL) / (sL * bL) * 1000)
    measures = {
        'shoulder':sL,
        'hip':hL, 
        'back':bL,
        'leg':lL
    }

    return score, measures


def video_SHA_score(vid, device, font_dir,  segment_model, segment_transform, keypoints_model, keypoints_transform):
    SHA_frames = []
    SHA_scores = []
    SHA_measures = []
    
    for index, frame in enumerate(vid):
        image = to_pil_image(frame)
        image = image.rotate(-90, expand=True)
        selected_box, keypoints = infer.detect_main_target(
            image, device, 0.8, keypoints_model, keypoints_transform
        )

        mask, mask_image, bool_mask = infer.segment_selected_target(
            image, device, selected_box, 0.99, segment_model, segment_transform
        )

        selected_kps = process.keypoints_filter(
             [
                'left_ear', 'right_ear',
                'left_shoulder','right_shoulder',
                'left_wrist','right_wrist',
                'left_hip', 'right_hip',
                'left_ankle', 'right_ankle'
            ], 
            keypoints
        )
        
        segment_area = process.segmentation_area(
            image, 
            bool_mask
        )
        
        hip_kps = process.find_hip_points(
            selected_kps['left_hip'], 
            selected_kps['right_hip'],
            selected_kps['left_wrist'],
            selected_kps['right_wrist'],
            segment_area
        )
        
        shoulder_kps = process.find_shoulder_points(
            selected_kps['left_shoulder'], 
            selected_kps['right_shoulder'],
            segment_area
        )

        if hip_kps == None or shoulder_kps == None:
            print("KEYPOINT ERRORS")
            continue
        
        main_points = {}
        main_points.update(hip_kps)
        main_points.update(shoulder_kps)
        main_points.update(
            {
                'left_ankle':selected_kps['left_ankle'],
                'right_ankle':selected_kps['right_ankle']
            }
        )
        
        middle_hip = measure.find_middle_point(
            main_points['left_hip'],
            main_points['right_hip']
        )

        float_labeled_frame = image
        for key, value in main_points.items():
            print(key, value)
            float_labeled_frame = draw.floating_rectangle_label(
                image = float_labeled_frame, 
                longitude_coordinate = middle_hip[0], 
                point=value, 
                label_text=key, 
                label_size=16, 
                label_color="#ffffff", 
                label_font=font_dir, 
                background_color="#11114A"
            )    

       
        score, measures = SHA_score(
            ls=main_points['left_shoulder'], 
            rs=main_points['right_shoulder'],
            lh=main_points['left_hip'],
            rh=main_points['right_hip'],
            la=main_points['left_ankle'],
            ra=main_points['right_ankle']
        )
        
        SHA_frames.append(float_labeled_frame)

        SHA_scores.append(score)
        SHA_measures.append(measures)

        print("##############################")
        print("Finished Processing ", index, " with score ", score, "\nand measures ", measures)
        print("##############################")
        
    return SHA_frames, SHA_scores, SHA_measures


def find_nearest(array, value):
	np_scores = np.array(array)
	distance_array = np.abs(np_scores - value)
	idx = distance_array.argmin()
	return array[idx], idx


def find_largest(array, value):
	np_scores = np.array(array)
	distance_array = np.abs(np_scores - value)
	idx = distance_array.argmax()
	return array[idx], idx


def best_scores(array, min_val, max_val):
    scores = []
    for x in array:
        if x > max_val or x <  min_val:
            pass
        else:
            scores.append(x)
            
    np_scores = np.array(scores)
    mean = np.mean(np_scores)
    median = np.median(np_scores)
    minim = np.min(np_scores)
    maxim = np.max(np_scores)
    histogram_scores = np.histogram(np_scores)
    print(mean, median)
    sns.distplot(np_scores, hist=True)
    
    return mean, median, minim, maxim







import os
import av 

from torchvision.models.detection import MaskRCNN_ResNet50_FPN_V2_Weights, maskrcnn_resnet50_fpn_v2
from torchvision.models.detection import keypointrcnn_resnet50_fpn, KeypointRCNN_ResNet50_FPN_Weights
from torchvision.io import read_video
from torchvision.transforms.functional import to_pil_image

### Load Human Segmentation Model
def segment_model(device):
	segment_weights = MaskRCNN_ResNet50_FPN_V2_Weights.COCO_V1
	segment_model = maskrcnn_resnet50_fpn_v2(weights=segment_weights)
	segment_transforms = segment_weights.transforms()
	segment_model.eval().to(device)
	return segment_model, segment_transforms


### Load Keypoints Detection Model
def keypoints_model(device):
	kp_weights = KeypointRCNN_ResNet50_FPN_Weights.COCO_V1
	kp_model =  keypointrcnn_resnet50_fpn(weights=kp_weights)
	kp_transforms = kp_weights.transforms()
	kp_model.eval().to(device)
	return kp_model, kp_transforms


### Load Video 
def video(video_path, rotate_angle, frame_position):
	video, audio, meta = read_video(
	    video_path,
	    pts_unit="sec",
	    output_format="TCHW"
	)

	frame_counts = len(video)
	fps = meta['video_fps']
	sample_index = frame_position
	sample_frame = to_pil_image(video[sample_index])
	sample_frame = sample_frame.rotate(rotate_angle, expand=True)
	return video, frame_counts, fps, sample_frame


### Load Images Directory
def image_frames(sample_dir):
	frames = sorted(os.listdir(sample_dir))
	total_frames = len(frames)
	#print("\nTotal Frames: ",total_frames)
	frames_path = ["" for x in range(total_frames)]
	#print("\nTotal Paths: ",len(frames_path))
	for frame_index, frame in enumerate(frames):     
		frames_path[frame_index] = os.path.join(sample_dir + "/" + frame)  
	return frames_path


from PIL import Image, ImageDraw, ImageFont, ImageFilter
from . import measure

def keypoint_markers(coordinates, image, label_font):
	marker_font = ImageFont.truetype(label_font, 8)
	sample = image.copy()
	draw = ImageDraw.Draw(sample) 
			
	for index, point in enumerate(coordinates.values()):
		label = str(index) 
		radius = 6

		print(point)

		draw.ellipse(
			[(point[0]-radius, point[1]-radius), (point[0]+radius, point[1]+radius)],
			fill="#40389F",
			outline="#ffffff",
			width=1
		)

		draw.text(
			(point[0]-3, point[1]-6),
			str(label),
			fill="#FAFF00",
			font=marker_font,
			align="center"
		)
	return sample


def circle_label(image, radius, central_point, background_color, label, label_font, label_color):
	point = central_point
	letter_length = len(label)
	label_size = int(radius/letter_length*2)
	marker_font = ImageFont.truetype(label_font, label_size)
	sample = image.copy()

	draw = ImageDraw.Draw(sample)
	draw.ellipse(
		[(point[0]-radius, point[1]-radius), (point[0]+radius, point[1]+radius)],
		fill=background_color,
		outline=background_color,
		width=1
	)

	draw.text(
		(point[0]-radius/2, point[1]-radius/1.8),
		str(label),
		fill=label_color,
		font=marker_font,
		align="center"
	) 

	return sample


def fixed_rectangle_label(image, anchor_point, label_text, label_size, label_font, label_color, background_color):
	letter_count = len(label_text)
	label_height = label_size * 2
	label_width = label_size * letter_count
	marker_font = ImageFont.truetype(label_font, label_size)
	sample = image.copy()

	rect_topleft = (
		(anchor_point[0] - label_width/2), 
		(anchor_point[1] - label_height/2)
	)

	rect_bottomright = (
		(anchor_point[0] + label_width/2), 
		(anchor_point[1] + label_height/2)
	)

	draw = ImageDraw.Draw(sample)

	draw.rectangle(
		[rect_topleft, rect_bottomright], 
		fill=background_color, 
		outline=label_color, 
		width=1
	)

	draw.text(
		(anchor_point[0] - label_width/4, anchor_point[1]-label_height/4),
		str(label_text),
		fill=label_color,
		font=marker_font,
		align="ms"
	)

	return sample 


def connecting_line(image, pointA, pointB, line_color, line_width):
	sample = image.copy()
	draw = ImageDraw.Draw(sample)
	draw.line(
		[pointA[0], pointA[1],  pointB[0], pointB[1]], 
		fill=line_color, 
		width=line_width
	)

	return sample


def floating_rectangle_label(image, longitude_coordinate, point, label_text, label_size, label_color, label_font, background_color):
	sample = image.copy()
	draw = ImageDraw.Draw(sample)
	radius = 6
	
	### Draw a Dot at anchor point
	draw.ellipse(
		[(point[0]-radius, point[1]-radius), (point[0]+radius, point[1]+radius)],
		fill=label_color,
		outline="#ffffff",
		width=1
	)

	### Draw a Label 60 degree North East or North West basing on anchor location
	marker_font = ImageFont.truetype(label_font, label_size)
	letter_count = len(label_text)
	label_height = label_size * 2
	label_width = label_size * letter_count

	if point[0] < longitude_coordinate:
		anchor_point = [point[0] - 100, point[1] - 100]
	else:
		anchor_point = [point[0] + 100, point[1] - 100]

	rect_topleft = (
		(anchor_point[0] - label_width/2), 
		(anchor_point[1] - label_height/2)
	)

	rect_bottomright = (
		(anchor_point[0] + label_width/2), 
		(anchor_point[1] + label_height/2)
	)

	rect_bottomleft = (
		rect_topleft[0], 
		rect_bottomright[1]
	)

	rect_topright = (
		rect_topleft[1],
		rect_bottomright[0]
	)

	draw.rectangle(
		[rect_topleft, rect_bottomright], 
		fill=background_color, 
		outline=label_color, 
		width=1
	)

	draw.text(
		(anchor_point[0] - label_width/4, anchor_point[1]-label_height/4),
		str(label_text),
		fill=label_color,
		font=marker_font,
		align="ms"
	)

	### Draw a Line connecting Anchor Dot and the Label
	line_color = label_color
	line_width = int(radius/3)

	draw.line(
		[point[0], point[1],  anchor_point[0], (anchor_point[1]+label_height/2)], 
		fill=line_color, 
		width=line_width
	)

	return sample 


def segmentation_contour(contour_pixels, contour_color, contour_size, font,  image):
	contoured = image.copy()
	for contour in contour_pixels:
	    contoured = circle_label(
	        contoured, 
	        contour_size, 
	        contour,
	        background_color=contour_color,
	        label=" ", 
	        label_font=font,
	        label_color=contour_color,   
	    )

	return contoured


def add_crown(score, le, re, top_head,frame, crown_image, font_file):
	middle_ear = measure.find_middle_point(le, re)
	head_width = measure.two_points_distance(le, re)
	crown_width = int(head_width * 1.6)
	crown_height = int(head_width * 0.8)
	crown_size = (crown_width,crown_height)
	crown = Image.open(crown_image)
	crown = crown.resize(crown_size)
	draw = ImageDraw.Draw(crown)
	font_size = int(crown_width/5)
	font = ImageFont.truetype(font_file, font_size)
	draw.text(
	    (crown_width/4,crown_height/2), 
	    str(score),
	    fill="#FF0000",
	    font=font,
	    align="center"
	)

	crown_position = (
	    int(top_head[0] - crown_width/2), 
	    int(top_head[1] - crown_height)
	)

	crowned = frame.copy()
	crowned.paste(crown, crown_position, crown)

	return crowned


# def add_thong(score, lh, rh, frame, font_file):
# 	return


def blur_background_around_bbox(bbox, sample_frame):
	top_left = (int(bbox[0]), int(bbox[1]))
	target = sample_frame.crop(bbox)
	blurImage = sample_frame.filter(ImageFilter.GaussianBlur(10))
	blurImage.paste(
		target, 
		top_left
	)
	return blurImage


def blur_background(segment_area, sample_frame):

	return sample_frame










import torch
import numpy as np
from torchvision.utils import draw_segmentation_masks
from torchvision.transforms.functional import to_pil_image, pil_to_tensor
from . import measure
from . import score

##########################################################
### Detect all Humans in Frames with High Probability ####
##########################################################
def detect_targets(frame, device, min_accuracy, kp_model, kp_transforms):

    ### Make Predictions on Frame
    body_matrix = kp_transforms(frame)
    predictions = kp_model([body_matrix.to(device)])

    ### Extract Boxes, Scores and Keypoints from frame
    boxes = predictions[0]['boxes']
    keypoints = predictions[0]['keypoints']
    scores = predictions[0]['scores']

    ### Filter for the main Bounding Box and Keypoints
    idx = torch.where(scores > min_accuracy)
    main_boxes = torch.squeeze(boxes[idx], dim=0)

    return main_boxes

###################################
### Detect Main Person in Frame ###
###################################
def detect_main_target(frame, device, min_accuracy, kp_model, kp_transforms):   

    ### Make Predictions on Frame
    body_matrix = kp_transforms(frame)
    predictions = kp_model([body_matrix.to(device)])

    ### Extract Boxes, Scores and Keypoints from frame
    boxes = predictions[0]['boxes']
    keypoints = predictions[0]['keypoints']
    scores = predictions[0]['scores']
    
    ### Filter for the main Bounding Box and Keypoints
    idx = torch.where(scores > min_accuracy)
    main_boxes = torch.unsqueeze(boxes[idx], dim=1)
    index = 0
    min_area = 0
    max_distance = frame.width/2

    print("Found ", len(main_boxes), " person in the frame")

    distances = []
    areas = []
    focuses = []
    bboxes = []
    
    for selector, box in enumerate(main_boxes):
        bbox = [
            box[0][0].item(), 
            box[0][1].item(),
            box[0][2].item(),
            box[0][3].item()
        ]
        
        distance, area = measure.distance_from_vertical_line(frame, bbox) 
        
        focuses.append(area/distance)
        distances.append(distance)
        areas.append(area)
        bboxes.append(bbox)

    print("Distances: ",distances)
    print("Areas: ", areas)
    print("FOCUSES: ", focuses)

    distance_score, distance_index = score.find_nearest(distances, 0)
    focus_score, focus_index = score.find_largest(focuses, 0)
    
    print("Select keypoint index", focus_index)
    selected_bbox = bboxes[focus_index] 
    selected_keypoints = torch.squeeze(keypoints[idx][focus_index], dim=0)
    # SHApoints = getSHAPositions(kp)
    # labels = ["ls","rs","lw", "rw", "lh","rh","la","ra"]
    # points_image = drawMarkers(labels, SHApoints, frame)
    return selected_bbox, selected_keypoints

###############################################
### Segment Selected Person from Background ###
###############################################
def segment_selected_target(frame, device, selected_bbox, segment_min_accuracy, segment_model, segment_transforms):
    ### Run Predicion Model to find Mask
    input_image = segment_transforms(frame)
    predictions = segment_model([input_image.to(device)])
    masks = predictions[0]['masks']
    scores = predictions[0]['scores']
    boxes = predictions[0]['boxes']
    idx = torch.where(scores > segment_min_accuracy)
    main_boxes = torch.unsqueeze(boxes[idx], dim=1)

    print("Found ", len(main_boxes), " person in the frame")
    # index = len(main_boxes) - 1
    
    distances = []
    for selector, box in enumerate(main_boxes):
        bbox = [
            box[0][0].item(), 
            box[0][1].item(),
            box[0][2].item(),
            box[0][3].item()
        ]
        
        distance_from_selected = measure.two_boxes_distance(bbox, selected_bbox)
        distances.append(distance_from_selected)
            
    print(distances)
    distance, index = score.find_nearest(distances, 0)
    print(distance, index)
    
    mask = torch.squeeze(masks[index][0], dim=1)
    mask_image = to_pil_image(mask)
    
    ### Find the Minimum Pixel Threshold for Mask
    numpy_mask = mask.detach().cpu().numpy()
    counts, values = np.histogram(numpy_mask, bins=10)
    min_counts = np.min(counts)
    idx = np.where(counts == min_counts)
    thresholds = sorted(values[idx]) 
    numpy_bool_mask = numpy_mask > thresholds[0]
    bool_mask = torch.tensor(numpy_bool_mask)
    
    return mask, mask_image, bool_mask


import torch
from PIL import ImageColor
from torchvision.transforms.functional import pil_to_tensor, to_pil_image
from . import score
from . import measure

# Keypoints_Filter
def keypoints_filter(selected_kpoints, detected_kpoints):
	kp_positions = {}
	coco_keypoints = [
		"nose", "left_eye", "right_eye", "left_ear", "right_ear",
		"left_shoulder", "right_shoulder", "left_elbow", "right_elbow",
		"left_wrist", "right_wrist", "left_hip", "right_hip",
		"left_knee", "right_knee", "left_ankle", "right_ankle",
	]

	for index, keypoint in enumerate(coco_keypoints):
		if keypoint in selected_kpoints:
			kp_positions[keypoint] = [
				detected_kpoints[index][0].item(), 
				detected_kpoints[index][1].item()
			] 
			# print(kp_positions[keypoint])

	return kp_positions


# Human Segmentation Area
def segmentation_area(sample_image, bool_mask):
	tensor_image = pil_to_tensor(sample_image)
	mask = torch.squeeze(bool_mask, 0)
	img_to_draw = tensor_image.detach().clone()
	color = ImageColor.getrgb('blue')
	tensor_color = torch.tensor(color, dtype=torch.uint8)
	img_to_draw[:, mask] = tensor_color[:, None]
	x, y = torch.where(
		(img_to_draw[2] == 255)
		&(img_to_draw[0] == 0)
		&(img_to_draw[1] == 0)
	)
	# print(x, y)
	positions = []
	for x, y in zip(y, x):
		positions.append(
			[x.item(), y.item()]
		)

	return positions


# Background Around Segmentation Area
def outside_segmentation_area(sample_image, bool_mask):
	inverse_bool_mask = ~bool_mask
	tensor_image = pil_to_tensor(sample_image)
	inverse_mask = torch.squeeze(inverse_bool_mask, 0)
	img_to_draw = tensor_image.detach().clone()
	color = ImageColor.getrgb('blue')
	tensor_color = torch.tensor(color, dtype=torch.uint8)
	img_to_draw[:, inverse_mask] = tensor_color[:, None]
	x, y = torch.where(
		(img_to_draw[2] == 255)
		&(img_to_draw[0] == 0)
		&(img_to_draw[1] == 0)
	)

	positions = []
	for x, y in zip(y, x):
	    positions.append(
	    	[x.item(), y.item()]
	    )
	return positions


# Human Segmentation Contour
def segmentation_contour(sample_image, bool_mask):
	tensor_image = pil_to_tensor(sample_image)
	mask = torch.squeeze(bool_mask, 0)
	img_to_draw = tensor_image.detach().clone()
	color = ImageColor.getrgb('blue')
	tensor_color = torch.tensor(color, dtype=torch.uint8)
	img_to_draw[:, mask] = tensor_color[:, None]
	colored = to_pil_image(img_to_draw)
	x, y = torch.where(
		(img_to_draw[2] == 255)
		&(img_to_draw[0] == 0)
		&(img_to_draw[1] == 0)
	)

	def check_if_contours(point, colored):
		[x, y] = point[0].item(), point[1].item()
		surrounding_pixels = [
			(x-1, y-1),(x, y-1),(x+1, y-1),
			(x-1, y), (x+1, y),
			(x-1, y+1), (x, y+1), (x+1, y+1)
		]
		checked = 0
		
		for pixel in surrounding_pixels:
			color = colored.getpixel(pixel)
			if color != (0, 0, 255):
				checked = checked + 1
				
		return checked

	contours = []
	for x, y in zip(y, x):
		result = check_if_contours(
			(x, y), 
			colored
		)
		
		if result != 0:
			contours.append([x.item(), y.item()])

	return contours


# Find Shoulder Points
def find_shoulder_points(ls, rs, segment_area):
	shoulder_alpha, shoulder_beta = score.two_points_linear_constant(ls, rs)

	if shoulder_alpha == None or shoulder_beta == None:
		return None

	shoulder_lines = score.find_segment_line(
	    segment_area, 
	    shoulder_alpha, 
	    shoulder_beta
	)

	if ls[0] < rs[0]:
		shoulder_kps = {
			'left_shoulder': shoulder_lines[0],
			'right_shoulder': shoulder_lines[-1]
		}
	elif ls[0] > rs[0]:
		shoulder_kps = {
			'left_shoulder': shoulder_lines[-1],
			'right_shoulder': shoulder_lines[0]
		}
	return shoulder_kps


# Find Hip Points with Hip Line Coordinate
def find_hip_points_HLC(lh, rh, lw, rw, segment_area):
	hip_alpha, hip_beta = score.two_points_linear_constant(lh, rh)

	if hip_alpha == None or hip_beta == None:
		return None

	hip_line_coordinates = score.find_segment_line(
		segment_area, 
		hip_alpha, 
		hip_beta
	)

	middle_hip = measure.find_middle_point(lh, rh)

	if rw[1] < rh[1] and lw[1] < lh[1]:
		hip_kps = {
			'left_hip': hip_line_coordinates[0],
			'right_hip': hip_line_coordinates[-1]
		}

		print(
			"both hand high", hip_kps,
			"\nLeft: ", lw, lh, 
			"\nRight: ", rw, rh
		)
		return hip_kps

	elif rw[1] > rh[1] and lw[1] > lh[1]:
		hip_kps = {}
		print(
			"both hand low", hip_kps,
			"\nLeft hand: ", lw[1], lh[1], 
			"\nRight: ", rw[1], rh[1]
		)

	elif int(lw[1]) >= int(lw[0] * hip_alpha + hip_beta)*0.9:
		precise_rh = hip_line_coordinates[-1]
		precise_lhX = 2 * middle_hip[0] - rh[0]
		precise_lhY = hip_alpha * precise_lhX + hip_beta
		precise_lh = [int(precise_lhX), int(precise_lhY)]
		hip_kps = {
			'left_hip': precise_lh,
			'right_hip': precise_rh
		}
		print(
			"low left hand", hip_kps, 
			"\n Left Wrist", lw,
			"\n Left Hip: ", lh)
		return hip_kps
		
		
	elif int(rw[1]) >= int(rw[0] * hip_alpha + hip_beta)*0.9:
		precise_lh = hip_line_coordinates[0]
		precise_rhX = 2 * middle_hip[0] - precise_lh[0]
		precise_rhY = hip_alpha * precise_rhX + hip_beta
		precise_rh = [int(precise_rhX), int(precise_rhY)] 
		
		hip_kps = {
			'left_hip': precise_lh,
			'right_hip': precise_rh
		}
		print(
			"low right hand", hip_kps,
			"\n Right Wrist: ", rw,
			"\n Right Hip: ", rh
		)
		return hip_kps


# Find Hip Points with Segmentation Area
def find_hip_points(lh, rh, lw, rw, segment_area):
	lhX = lh[0]
	lhY = lh[1]

	rhX = rh[0]
	rhY = rh[1]

	hip_alpha, hip_beta = score.two_points_linear_constant(lh, rh)

	if hip_alpha == None or hip_beta == None:
		return None

	hip_line_coordinates = score.find_segment_line(
		segment_area, 
		hip_alpha, 
		hip_beta
	)

	middle_hip = measure.find_middle_point(lh, rh)
	lThreshold = middle_hip[0] - measure.two_points_distance(
		measure.find_middle_point(lh, rh), lh
	) * 1.8

	rThreshold = middle_hip[0] + measure.two_points_distance(
		measure.find_middle_point(lh, rh), rh
	) * 1.8

	for idx, position in enumerate(segment_area):
		expectedY = hip_alpha * position[0] + hip_beta

		if int(position[1]) == int(expectedY) and position[0] > lThreshold and position[0] < middle_hip[0]:
			lhY = position[1]

			if position[0] < lhX:
				lhX = position[0]

		elif int(position[1]) == int(expectedY) and position[0] < rThreshold and position[0] > middle_hip[0]:
			rhY = position[1]
			if position[0] > rhX:
				rhX = position[0]

	if lh[0] < rh[0]:
		hip_kps = {
			'left_hip': (lhX, lhY), 
			'right_hip': (rhX, rhY)
		}

		return hip_kps

	elif  lh[0] > rh[0]:
		hip_kps = {
			'left_hip': (rhX, rhY), 
			'right_hip': (lhX, lhY)
		}
		return hip_kps
		

# Find Top Head Points
def find_tophead_point(le, re, segment_contours):
	tophead_pt = {}
	middle_ear = measure.find_middle_point(le, re)

	for contour_point in segment_contours:
	    if int(contour_point[0]) == int(middle_ear[0]) and int(contour_point[1]) < int(middle_ear[1]):
	        tophead_pt = (contour_point[0], contour_point[1])
	        
	return tophead_pt







import av
from PIL import Image

##############################################
##### Generate Video from PIL Image Arrays ###
##############################################
def generate_video_from_pil_images(pil_images, output, width, height):
	container = av.open(output, mode="w")

	stream = container.add_stream('mpeg4', rate=30)
	stream.width = width
	stream.height = height

	def encode(pil_image):
		frame = av.VideoFrame.from_image(pil_image)
		for packet in stream.encode(frame):
			container.mux(packet)
		
	for pil_image in pil_images:
		encode(pil_image)
	
	container.close()


##############################
#### Generate Seek Video #####
##############################
def generate_seek_video(pil_images, target_index, output, width, height):
	container = av.open(output, mode="w")
	
	stream = container.add_stream("mpeg4", rate=30)
	stream.width = width
	stream.height = height
	
	def encode(pil_image):
		frame = av.VideoFrame.from_image(pil_image)
		for packet in stream.encode(frame):
			container.mux(packet)
			
	reversed_images = pil_images[::-1]
	second_part = pil_images[target_index:-1] 
	reversed_second_part = second_part[::-1]
	
	seek_images = pil_images + reversed_images + pil_images + reversed_second_part
	focus_images = [pil_images[target_index]] * 180
	
	for image in seek_images :
		slow_motion_frames = [image] * 6
		for frame in slow_motion_frames:
			encode(frame)
	
	for focus_image in focus_images:
		encode(focus_image)
		
	container.close()


###################################
##### Generate Instagram Video ####
###################################
def generate_instagram_vid(vid_name, vid_width, vid_height, pil_images, stop_index, fps, repeat_rate,slow_motion_rate):
	container = av.open(vid_name, mode="w")

	### Add mp4 stream to container with fps, width and height
	stream = container.add_stream("mpeg4", rate=int(fps))
	stream.width = vid_width
	stream.height = vid_height
	
	### Write Encode function to Encode frame into Mpeg4 format
	def encode(image):
		frame = av.VideoFrame.from_image(image)
		for packet in stream.encode(frame):
			container.mux(packet)
			
	def slow_motion_reverse(pil_images):
		frames_count = len(pil_images)
		total_frames = frames_count * 2 * repeat_rate
		print("Total Frames is ", total_frames)
		
		for idx in range(total_frames):
			local_index = idx%len(pil_images)
			play_times = int(idx/len(pil_images))
		
			if play_times % 2 == 0 and play_times < (repeat_rate * 2 - 1):
				frame_index = local_index
				print(frame_index, play_times)
				frame = pil_images[local_index]
				frames = [frame] * slow_motion_rate
				for frame in frames:
					encode(frame)
					
			elif play_times == (repeat_rate * 2 - 1):
				frame_index = frames_count - local_index - 1
				if frame_index > stop_index:
					print(frame_index, play_times)
					frame = pil_images[frame_index]
					frames = [frame] * slow_motion_rate * 3
					for frame in frames:
						encode(frame)

				elif frame_index == stop_index:
					print(frame_index, play_times)
					frame = pil_images[frame_index]
					frames = [frame] * slow_motion_rate * 90
					for frame in frames:
						encode(frame)
								
			else:
				print(frame_index, play_times)
				frame = pil_images[frames_count - local_index - 1]
				frames = [frame] * slow_motion_rate
				for frame in frames:
					encode(frame)
			
	slow_motion_reverse(pil_images)

	### Close Container
	container.close()


###########################################
##### Generate Long Video from Frames  ####
###########################################
def generate_long_video(pil_images, scroll_speed, fps):
    total_frames = len(pil_images)
    image = pil_images[0]
    long_image_width = image.width * total_frames 
    long_image_height = image.height
    long_image =  Image.new("RGB", (long_image_width, long_image_height))
    
    for idx in range(total_frames):
        x = idx * image.width
        image = pil_images[idx]
        long_image.paste(image, box=(x, 0))
        if idx == total_frames - 1:
            long_image.save("long_image.png")
    
    stream_duration = int(long_image_width/scroll_speed)
    video_name = str(stream_duration)+"_"+str(scroll_speed)+"_"+str(fps)+"_scroll_video.mp4"
        
    container = av.open(video_name, mode="w")
    stream = container.add_stream("mpeg4", rate=int(fps))
    stream.width = int(image.height * 16 / 9)
    stream.height = image.height
    
    def encode(image):
        frame = av.VideoFrame.from_image(image)
        for packet in stream.encode(frame):
            container.mux(packet)
            
    for idx in range(stream_duration):
        new_frame = Image.new("RGB", (stream.width, stream.height))
        x = 0 - idx*scroll_speed
        print(idx, stream_duration)
        
        if x <= (stream.width - long_image_width):
            freeze_x = stream.width - long_image_width
            new_frame.paste(im=long_image, box=(freeze_x,0))
            encode(new_frame)
        else:
            new_frame.paste(im=long_image, box=(x,0))
            encode(new_frame)
        
    container.close()


